#include <stdio.h>
#include <math.h>

/* The bisection method is meant to compute the zeros of 
   a continuous function f in an interval [a,b] in which 
   f(a)*f(b) < 0. So, in order to compute the square root 
   of a number, we need to implement a function whose 
   zeros correspond to the root of a parameter a: this
   function is root_square_delta */
double root_square_delta(const double a, const double x)
{
   return x*x-a;
}

/* This is a general bisection method for functions whose 
   signature is f(const double a, const double x) where 
   a is a parameter (e.g., root_square_delta) */
double get_delta_root(double (*f)(const double, const double),
                      double a,
		      double lb, 
		      double rb,
		      const double delta)
{
   double m=(lb+rb)/2;

   double f_lb=f(a,lb),
	  f_m=f(a,m);

   while (fabs(f_m)>=delta && 
          m!=lb     /* This further condition is mean to deal with 
                       the finite arithmetic of the computer:
                       if f(a,x) is too steep in a neighborhood 
                       of its root, when it may be the case that 
                       the computation reduces the interval 
                       to one single point (i.e., lb=rb), but 
                       |f(a, lb)| is still greater than delta. 
                       In these cases, lb is the best 
                       approximation for the root among the 
                       doubles */
         ) {
       if (f_lb*f_m<0) {
	 rb=m;
       } else {
	 lb=m;
         f_lb=f_m;	 
       }

       m=(lb+rb)/2;
       f_m=f(a,m);
   }

   return m;
}


double square_root(const double a, const double delta)
{
   /* The inteval to be investigated to compute the 
      square root of a value depends on the value itself:
      if the value is smaller than 1, then we must 
      consider the interval [0,1];
      if the value is greater than or equal to 1, then, we must 
      take care of the interval [1, a] */

   if (a<1.0) {
	return get_delta_root(root_square_delta, a, 0.0, 1.0, delta);
   }

   if (a>1.0) {
	return get_delta_root(root_square_delta, a, 1.0, a, delta);
   }

   return 1.0;
}

int main(int argc, char *argv[])
{
   if (argc<2) {
     fprintf(stderr, "Syntax error: expected one parameter\n%s <floating point number>\n",          
             argv[0]);
     return 1;
   }
	
   double a;
   sscanf(argv[1], "%lf", &a);   

   double x = square_root(a, 1e-10);

   printf("%lf is a root for f(x)=x*x-%lf:\n%lf^2=%lf\n%lf^2-%lf^2=%.3e\n", 
          x, a, x, x*x, x, a, x*x-a);

   return 0;
}
